import asyncio
import binascii
import contextlib
import random
import socket
from unittest import TestCase, skipIf
from unittest.mock import patch

from cryptography.hazmat.primitives import serialization
from qh3.asyncio.client import connect
from qh3.asyncio.protocol import QuicConnectionProtocol
from qh3.asyncio.server import serve
from qh3.quic.configuration import QuicConfiguration
from qh3.quic.logger import QuicLogger

from .utils import (
    SERVER_CACERTFILE,
    SERVER_CERTFILE,
    SERVER_COMBINEDFILE,
    SERVER_KEYFILE,
    SKIP_TESTS,
    asynctest,
    generate_ec_certificate,
    generate_ed448_certificate,
    generate_ed25519_certificate,
)

real_sendto = socket.socket.sendto


def sendto_with_loss(self, data, addr=None):
    """
    Simulate 25% packet loss.
    """
    if random.random() > 0.25:
        real_sendto(self, data, addr)


class SessionTicketStore:
    def __init__(self):
        self.tickets = {}

    def add(self, ticket):
        self.tickets[ticket.ticket] = ticket

    def pop(self, label):
        return self.tickets.pop(label, None)


def handle_stream(reader, writer):
    async def serve():
        data = await reader.read()
        writer.write(bytes(reversed(data)))
        writer.write_eof()

    asyncio.ensure_future(serve())


class HighLevelTest(TestCase):
    def setUp(self):
        self.bogus_port = 1024
        self.server_host = "localhost"

    async def run_client(
        self,
        *,
        port: int,
        host=None,
        cadata=None,
        cafile=SERVER_CACERTFILE,
        configuration=None,
        request=b"ping",
        **kwargs,
    ):
        if host is None:
            host = self.server_host
        if configuration is None:
            configuration = QuicConfiguration(is_client=True)
        configuration.load_verify_locations(cadata=cadata, cafile=cafile)
        async with connect(host, port, configuration=configuration, **kwargs) as client:
            # waiting for connected when connected returns immediately
            await client.wait_connected()

            reader, writer = await client.create_stream()
            self.assertEqual(writer.can_write_eof(), True)
            self.assertEqual(writer.get_extra_info("stream_id"), 0)

            writer.write(request)
            writer.write_eof()

            response = await reader.read()

        # waiting for closed when closed returns immediately
        await client.wait_closed()

        return response

    @contextlib.asynccontextmanager
    async def run_server(self, configuration=None, host="::", **kwargs):
        if configuration is None:
            configuration = QuicConfiguration(is_client=False)
            configuration.load_cert_chain(SERVER_CERTFILE, SERVER_KEYFILE)
        server = await serve(
            host=host,
            port=0,
            configuration=configuration,
            stream_handler=handle_stream,
            **kwargs,
        )
        try:
            yield server._transport.get_extra_info("sockname")[1]
        finally:
            server.close()

    @asynctest
    async def test_connect_and_serve(self):
        async with self.run_server() as server_port:
            response = await self.run_client(port=server_port)
            self.assertEqual(response, b"gnip")

    @asynctest
    async def test_connect_and_serve_ipv4(self):
        async with self.run_server(host="0.0.0.0") as server_port:
            response = await self.run_client(host="127.0.0.1", port=server_port)
            self.assertEqual(response, b"gnip")

    @skipIf("ipv6" in SKIP_TESTS, "Skipping IPv6 tests")
    @asynctest
    async def test_connect_and_serve_ipv6(self):
        async with self.run_server(host="::") as server_port:
            response = await self.run_client(host="::1", port=server_port)
            self.assertEqual(response, b"gnip")

    async def _test_connect_and_serve_with_certificate(self, certificate, private_key):
        async with self.run_server(
            configuration=QuicConfiguration(
                certificate=certificate,
                private_key=private_key,
                is_client=False,
            )
        ) as server_port:
            response = await self.run_client(
                cadata=certificate.public_bytes(serialization.Encoding.PEM),
                cafile=None,
                port=server_port,
            )
            self.assertEqual(response, b"gnip")

    @asynctest
    async def test_connect_and_serve_with_ec_certificate(self):
        await self._test_connect_and_serve_with_certificate(
            *generate_ec_certificate(common_name="localhost")
        )

    @asynctest
    async def test_connect_and_serve_with_ed25519_certificate(self):
        await self._test_connect_and_serve_with_certificate(
            *generate_ed25519_certificate(common_name="localhost")
        )

    @asynctest
    async def test_connect_and_serve_with_ed448_certificate(self):
        await self._test_connect_and_serve_with_certificate(
            *generate_ed448_certificate(common_name="localhost")
        )

    @asynctest
    async def test_connect_and_serve_large(self):
        """
        Transfer enough data to require raising MAX_DATA and MAX_STREAM_DATA.
        """
        data = b"Z" * 2097152
        async with self.run_server() as server_port:
            response = await self.run_client(port=server_port, request=data)
            self.assertEqual(response, data)

    @asynctest
    async def test_connect_and_serve_without_client_configuration(self):
        async with self.run_server() as server_port:
            with self.assertRaises(ConnectionError):
                async with connect(self.server_host, server_port) as client:
                    await client.ping()

    @asynctest
    async def test_connect_and_serve_writelines(self):
        async with self.run_server() as server_port:
            configuration = QuicConfiguration(is_client=True)
            configuration.load_verify_locations(cafile=SERVER_CACERTFILE)
            async with connect(
                self.server_host, server_port, configuration=configuration
            ) as client:
                reader, writer = await client.create_stream()
                assert writer.can_write_eof() is True

                writer.writelines([b"01234567", b"89012345"])
                writer.write_eof()

                response = await reader.read()
                self.assertEqual(response, b"5432109876543210")

    @skipIf("loss" in SKIP_TESTS, "Skipping loss tests")
    @patch("socket.socket.sendto", new_callable=lambda: sendto_with_loss)
    @asynctest
    async def test_connect_and_serve_with_packet_loss(self, mock_sendto):
        """
        This test ensures handshake success and stream data is successfully sent
        and received in the presence of packet loss (randomized 25% in each direction).
        """
        data = b"Z" * 65536

        server_configuration = QuicConfiguration(
            is_client=False, quic_logger=QuicLogger()
        )
        server_configuration.load_cert_chain(SERVER_CERTFILE, SERVER_KEYFILE)
        async with self.run_server(configuration=server_configuration) as server_port:
            response = await self.run_client(
                configuration=QuicConfiguration(
                    is_client=True, quic_logger=QuicLogger()
                ),
                port=server_port,
                request=data,
            )
        self.assertEqual(response, data)

    @asynctest
    async def test_connect_and_serve_with_session_ticket(self):
        client_ticket = None
        store = SessionTicketStore()

        def save_ticket(t):
            nonlocal client_ticket
            client_ticket = t

        async with self.run_server(
            session_ticket_fetcher=store.pop, session_ticket_handler=store.add
        ) as server_port:
            # first request
            response = await self.run_client(
                port=server_port, session_ticket_handler=save_ticket
            )
            self.assertEqual(response, b"gnip")

            self.assertIsNotNone(client_ticket)

            # second request
            response = await self.run_client(
                configuration=QuicConfiguration(
                    is_client=True, session_ticket=client_ticket
                ),
                port=server_port,
            )
            self.assertEqual(response, b"gnip")

    @asynctest
    async def test_connect_and_serve_with_retry(self):
        async with self.run_server(retry=True) as server_port:
            response = await self.run_client(port=server_port)
            self.assertEqual(response, b"gnip")

    @asynctest
    async def test_connect_and_serve_with_retry_bad_original_destination_connection_id(
        self,
    ):
        """
        If the server's transport parameters do not have the correct
        original_destination_connection_id the connection must fail.
        """

        def create_protocol(*args, **kwargs):
            protocol = QuicConnectionProtocol(*args, **kwargs)
            protocol._quic._original_destination_connection_id = None
            return protocol

        async with self.run_server(
            create_protocol=create_protocol, retry=True
        ) as server_port:
            with self.assertRaises(ConnectionError):
                await self.run_client(port=server_port)

    @asynctest
    async def test_connect_and_serve_with_retry_bad_retry_source_connection_id(self):
        """
        If the server's transport parameters do not have the correct
        retry_source_connection_id the connection must fail.
        """

        def create_protocol(*args, **kwargs):
            protocol = QuicConnectionProtocol(*args, **kwargs)
            protocol._quic._retry_source_connection_id = None
            return protocol

        async with self.run_server(
            create_protocol=create_protocol, retry=True
        ) as server_port:
            with self.assertRaises(ConnectionError):
                await self.run_client(port=server_port)

    @patch("qh3.quic.retry.QuicRetryTokenHandler.validate_token")
    @asynctest
    async def test_connect_and_serve_with_retry_bad_token(self, mock_validate):
        mock_validate.side_effect = ValueError("Decryption failed.")

        async with self.run_server(retry=True) as server_port:
            with self.assertRaises(ConnectionError):
                await self.run_client(
                    configuration=QuicConfiguration(is_client=True, idle_timeout=4.0),
                    port=server_port,
                )

    @asynctest
    async def test_connect_and_serve_with_version_negotiation(self):
        async with self.run_server() as server_port:
            # force version negotiation
            configuration = QuicConfiguration(is_client=True, quic_logger=QuicLogger())
            configuration.supported_versions.insert(0, 0x1A2A3A4A)

            response = await self.run_client(
                configuration=configuration, port=server_port
            )
            self.assertEqual(response, b"gnip")

    @asynctest
    async def test_connect_timeout(self):
        with self.assertRaises(ConnectionError):
            await self.run_client(
                port=self.bogus_port,
                configuration=QuicConfiguration(is_client=True, idle_timeout=5),
            )

    @asynctest
    async def test_connect_timeout_no_wait_connected(self):
        with self.assertRaises(ConnectionError):
            configuration = QuicConfiguration(is_client=True, idle_timeout=5)
            configuration.load_verify_locations(cafile=SERVER_CACERTFILE)
            async with connect(
                self.server_host,
                self.bogus_port,
                configuration=configuration,
                wait_connected=False,
            ) as client:
                await client.ping()

    @asynctest
    async def test_connect_local_port(self):
        async with self.run_server() as server_port:
            response = await self.run_client(local_port=3456, port=server_port)
            self.assertEqual(response, b"gnip")

    @asynctest
    async def test_connect_local_port_bind(self):
        with self.assertRaises(OverflowError):
            await self.run_client(local_port=-1, port=self.bogus_port)

    @asynctest
    async def test_change_connection_id(self):
        async with self.run_server() as server_port:
            configuration = QuicConfiguration(is_client=True)
            configuration.load_verify_locations(cafile=SERVER_CACERTFILE)
            async with connect(
                self.server_host, server_port, configuration=configuration
            ) as client:
                await client.ping()
                client.change_connection_id()
                await client.ping()

    @asynctest
    async def test_key_update(self):
        async with self.run_server() as server_port:
            configuration = QuicConfiguration(is_client=True)
            configuration.load_verify_locations(cafile=SERVER_CACERTFILE)
            async with connect(
                self.server_host, server_port, configuration=configuration
            ) as client:
                await client.ping()
                client.request_key_update()
                await client.ping()

    @asynctest
    async def test_ping(self):
        async with self.run_server() as server_port:
            configuration = QuicConfiguration(is_client=True)
            configuration.load_verify_locations(cafile=SERVER_CACERTFILE)
            async with connect(
                self.server_host, server_port, configuration=configuration
            ) as client:
                await client.ping()
                await client.ping()

    @asynctest
    async def test_ping_parallel(self):
        async with self.run_server() as server_port:
            configuration = QuicConfiguration(is_client=True)
            configuration.load_verify_locations(cafile=SERVER_CACERTFILE)
            async with connect(
                self.server_host, server_port, configuration=configuration
            ) as client:
                coros = [client.ping() for x in range(16)]
                await asyncio.gather(*coros)

    @asynctest
    async def test_server_receives_garbage(self):
        configuration = QuicConfiguration(is_client=False)
        configuration.load_cert_chain(SERVER_CERTFILE, SERVER_KEYFILE)
        server = await serve(
            host=self.server_host,
            port=0,
            configuration=configuration,
        )
        server.datagram_received(binascii.unhexlify("c00000000080"), ("1.2.3.4", 1234))
        server.close()

    @asynctest
    async def test_combined_key(self):
        config1 = QuicConfiguration()
        config2 = QuicConfiguration()
        config1.load_cert_chain(SERVER_CERTFILE, SERVER_KEYFILE)
        config2.load_cert_chain(SERVER_COMBINEDFILE)

        self.assertEqual(config1.certificate, config2.certificate)



from typing import Tuple

from cryptography.hazmat.bindings.openssl.binding import Binding

AEAD_KEY_LENGTH_MAX = 32
AEAD_NONCE_LENGTH = 12
AEAD_TAG_LENGTH = 16

PACKET_LENGTH_MAX = 1500
SAMPLE_LENGTH = 16
PACKET_NUMBER_LENGTH_MAX = 4


class CryptoError(ValueError):
    pass


def _get_cipher_by_name(binding: Binding, cipher_name: bytes):  # -> EVP_CIPHER
    evp_cipher = binding.lib.EVP_get_cipherbyname(cipher_name)
    if evp_cipher == binding.ffi.NULL:
        raise CryptoError(f"Invalid cipher name: {cipher_name.decode()}")
    return evp_cipher


class _CryptoBase:
    def __init__(self) -> None:
        self._binding = Binding()

    def _handle_openssl_failure(self) -> bool:
        self._binding.lib.ERR_clear_error()
        raise CryptoError("OpenSSL call failed")


class AEAD(_CryptoBase):
    def __init__(self, cipher_name: bytes, key: bytes, iv: bytes) -> None:
        super().__init__()

        # check and store key and iv
        if len(key) > AEAD_KEY_LENGTH_MAX:
            raise CryptoError("Invalid key length")
        self._key = key
        if len(iv) != AEAD_NONCE_LENGTH:
            raise CryptoError("Invalid iv length")
        self._iv = iv

        # create cipher contexts
        evp_cipher = _get_cipher_by_name(self._binding, cipher_name)
        self._decrypt_ctx = self._create_ctx(evp_cipher, operation=0)
        self._encrypt_ctx = self._create_ctx(evp_cipher, operation=1)

        # allocate buffers
        self._nonce = self._binding.ffi.new("unsigned char[]", AEAD_NONCE_LENGTH)
        self._buffer = self._binding.ffi.new("unsigned char[]", PACKET_LENGTH_MAX)
        self._buffer_view = self._binding.ffi.buffer(self._buffer)
        self._outlen = self._binding.ffi.new("int *")
        self._dummy_outlen = self._binding.ffi.new("int *")

    def _create_ctx(self, evp_cipher, operation: int):  # -> EVP_CIPHER_CTX *
        # create a cipher context with the given type and operation mode
        ctx = self._binding.ffi.gc(
            self._binding.lib.EVP_CIPHER_CTX_new(),
            self._binding.lib.EVP_CIPHER_CTX_free,
        )
        ctx != self._binding.ffi.NULL or self._handle_openssl_failure()
        self._binding.lib.EVP_CipherInit_ex(
            ctx,  # EVP_CIPHER_CTX *ctx
            evp_cipher,  # const EVP_CIPHER *type
            self._binding.ffi.NULL,  # ENGINE *impl
            self._binding.ffi.NULL,  # const unsigned char *key
            self._binding.ffi.NULL,  # const unsigned char *iv
            operation,  # int enc
        ) == 1 or self._handle_openssl_failure()

        # specify key and initialization vector length
        self._binding.lib.EVP_CIPHER_CTX_set_key_length(
            ctx,  # EVP_CIPHER_CTX *ctx
            len(self._key),  # int keylen
        ) == 1 or self._handle_openssl_failure()
        self._binding.lib.EVP_CIPHER_CTX_ctrl(
            ctx,  # EVP_CIPHER_CTX *ctx
            self._binding.lib.EVP_CTRL_AEAD_SET_IVLEN,  # int cmd
            AEAD_NONCE_LENGTH,  # int ivlen
            self._binding.ffi.NULL,  # void *NULL
        ) == 1 or self._handle_openssl_failure()
        return ctx

    def _init_nonce(self, packet_number: int) -> None:
        # reference: https://datatracker.ietf.org/doc/html/rfc9001#section-5.3

        # left-pad the reconstructed packet number (62 bits ~ 8 bytes)
        # and XOR it with the IV
        self._binding.ffi.memmove(self._nonce, self._iv, AEAD_NONCE_LENGTH)
        for i in range(8):
            if packet_number == 0:
                break
            self._nonce[AEAD_NONCE_LENGTH - 1 - i] ^= packet_number & 0xFF
            packet_number >>= 8

    def decrypt(self, data: bytes, associated_data: bytes, packet_number: int) -> bytes:
        if len(data) < AEAD_TAG_LENGTH or len(data) > PACKET_LENGTH_MAX:
            raise CryptoError("Invalid payload length")
        self._init_nonce(packet_number)

        # get the appended AEAD tag (data = cipher text + tag)
        cipher_text_len = len(data) - AEAD_TAG_LENGTH
        self._binding.lib.EVP_CIPHER_CTX_ctrl(
            self._decrypt_ctx,  # EVP_CIPHER_CTX *ctx
            self._binding.lib.EVP_CTRL_AEAD_SET_TAG,  # int cmd
            AEAD_TAG_LENGTH,  # int taglen
            data[cipher_text_len:],  # void *tag
        ) == 1 or self._handle_openssl_failure()

        # set key and nonce
        self._binding.lib.EVP_CipherInit_ex(
            self._decrypt_ctx,  # EVP_CIPHER_CTX *ctx
            self._binding.ffi.NULL,  # const EVP_CIPHER *type
            self._binding.ffi.NULL,  # ENGINE *impl
            self._key,  # const unsigned char *key
            self._nonce,  # const unsigned char *iv
            0,  # int enc
        ) == 1 or self._handle_openssl_failure()

        # specify the header as additional authenticated data (AAD)
        self._binding.lib.EVP_CipherUpdate(
            self._decrypt_ctx,  # EVP_CIPHER_CTX *ctx
            self._binding.ffi.NULL,  # unsigned char *out
            self._dummy_outlen,  # int *outl
            associated_data,  # const unsigned char *in
            len(associated_data),  # int inl
        ) == 1 or self._handle_openssl_failure()

        # decrypt the cipher text (i.e. received data excluding the appended tag)
        self._binding.lib.EVP_CipherUpdate(
            self._decrypt_ctx,  # EVP_CIPHER_CTX *ctx
            self._buffer,  # unsigned char *out
            self._outlen,  # int *outl
            data,  # const unsigned char *in
            cipher_text_len,  # int inl
        ) == 1 or self._handle_openssl_failure()

        # finalize the operation
        self._binding.lib.EVP_CipherFinal_ex(
            self._decrypt_ctx,  # EVP_CIPHER_CTX *ctx
            self._binding.ffi.NULL,  # unsigned char *outm
            self._dummy_outlen,  # int *outl
        ) == 1 or self._handle_openssl_failure()

        # return the decrypted data
        return self._buffer_view[: self._outlen[0]]

    def encrypt(self, data: bytes, associated_data: bytes, packet_number: int) -> bytes:
        if len(data) > PACKET_LENGTH_MAX:
            raise CryptoError("Invalid payload length")
        self._init_nonce(packet_number)

        # set key and nonce
        self._binding.lib.EVP_CipherInit_ex(
            self._encrypt_ctx,  # EVP_CIPHER_CTX *ctx
            self._binding.ffi.NULL,  # const EVP_CIPHER *type
            self._binding.ffi.NULL,  # ENGINE *impl
            self._key,  # const unsigned char *key
            self._nonce,  # const unsigned char *iv
            1,  # int enc
        ) == 1 or self._handle_openssl_failure()

        # specify the header as additional authenticated data (AAD)
        self._binding.lib.EVP_CipherUpdate(
            self._encrypt_ctx,  # EVP_CIPHER_CTX *ctx
            self._binding.ffi.NULL,  # unsigned char *out
            self._dummy_outlen,  # int *outl
            associated_data,  # const unsigned char *in
            len(associated_data),  # int inl
        ) == 1 or self._handle_openssl_failure()

        # encrypt the data
        self._binding.lib.EVP_CipherUpdate(
            self._encrypt_ctx,  # EVP_CIPHER_CTX *ctx
            self._buffer,  # unsigned char *out
            self._outlen,  # int *outl
            data,  # const unsigned char *in
            len(data),  # int inl
        ) == 1 or self._handle_openssl_failure()

        # finalize the operation
        self._binding.lib.EVP_CipherFinal_ex(
            self._encrypt_ctx,  # EVP_CIPHER_CTX *ctx
            self._binding.ffi.NULL,  # unsigned char *outm
            self._dummy_outlen,  # int *outl
        ) == 1 and self._dummy_outlen[0] == 0 or self._handle_openssl_failure()

        # append the AEAD tag to the cipher text
        outlen_with_tag = self._outlen[0] + AEAD_TAG_LENGTH
        if outlen_with_tag > PACKET_LENGTH_MAX:
            raise CryptoError("Invalid payload length")
        self._binding.lib.EVP_CIPHER_CTX_ctrl(
            self._encrypt_ctx,  # EVP_CIPHER_CTX *ctx
            self._binding.lib.EVP_CTRL_AEAD_GET_TAG,  # int cmd
            AEAD_TAG_LENGTH,  # int taglen
            self._buffer + self._outlen[0],  # void *tag
        ) == 1 or self._handle_openssl_failure()

        # return the encrypted cipher text and AEAD tag
        return self._buffer_view[:outlen_with_tag]


class HeaderProtection(_CryptoBase):
    def __init__(self, cipher_name: bytes, key: bytes) -> None:
        super().__init__()
        self._is_chacha20 = cipher_name == b"chacha20"
        if len(key) > AEAD_KEY_LENGTH_MAX:
            raise CryptoError("Invalid key length")

        # create cipher with given type
        evp_cipher = _get_cipher_by_name(self._binding, cipher_name)
        self._ctx = self._binding.ffi.gc(
            self._binding.lib.EVP_CIPHER_CTX_new(),
            self._binding.lib.EVP_CIPHER_CTX_free,
        )
        self._ctx != self._binding.ffi.NULL or self._handle_openssl_failure()
        self._binding.lib.EVP_CipherInit_ex(
            self._ctx,  # EVP_CIPHER_CTX *ctx
            evp_cipher,  # const EVP_CIPHER *type
            self._binding.ffi.NULL,  # ENGINE *impl
            self._binding.ffi.NULL,  # const unsigned char *key
            self._binding.ffi.NULL,  # const unsigned char *iv
            1,  # int enc
        ) == 1 or self._handle_openssl_failure()

        # set cipher key
        self._binding.lib.EVP_CIPHER_CTX_set_key_length(
            self._ctx,  # EVP_CIPHER_CTX *ctx
            len(key),  # int keylen
        ) == 1 or self._handle_openssl_failure()
        self._binding.lib.EVP_CipherInit_ex(
            self._ctx,  # EVP_CIPHER_CTX *ctx
            self._binding.ffi.NULL,  # const EVP_CIPHER *type
            self._binding.ffi.NULL,  # ENGINE *impl
            key,  # const unsigned char *key
            self._binding.ffi.NULL,  # const unsigned char *iv
            1,  # int enc
        ) == 1 or self._handle_openssl_failure()

        # allocate buffers
        self._buffer = self._binding.ffi.new("unsigned char[]", PACKET_LENGTH_MAX)
        self._buffer_view = self._binding.ffi.buffer(self._buffer)
        self._dummy_outlen = self._binding.ffi.new("int *")
        self._mask = self._binding.ffi.new("unsigned char[]", 31)
        self._zero = self._binding.ffi.new("unsigned char[]", 5)

    def _update_mask(self, pn_offset: int, buffer_len: int) -> None:
        # reference: https://datatracker.ietf.org/doc/html/rfc9001#section-5.4.2

        # sample data starts 4 bytes after the beginning of the Packet Number field
        # (regardless of its length)
        sample_offset = pn_offset + 4
        assert pn_offset + SAMPLE_LENGTH <= buffer_len

        if self._is_chacha20:
            # reference: https://datatracker.ietf.org/doc/html/rfc9001#section-5.4.4

            # the first four bytes after pn_offset are block counter,
            # the next 12 bytes are the nonce
            self._binding.lib.EVP_CipherInit_ex(
                self._ctx,  # EVP_CIPHER_CTX *ctx
                self._binding.ffi.NULL,  # const EVP_CIPHER *type
                self._binding.ffi.NULL,  # ENGINE *impl
                self._binding.ffi.NULL,  # const unsigned char *key
                self._buffer + sample_offset,  # const unsigned char *iv
                1,  # int enc
            ) == 1 or self._handle_openssl_failure()

            # ChaCha20 is used to protect 5 zero bytes
            self._binding.lib.EVP_CipherUpdate(
                self._ctx,  # EVP_CIPHER_CTX *ctx
                self._mask,  # unsigned char *out
                self._dummy_outlen,  # int *outl
                self._zero,  # const unsigned char *in
                len(self._zero),  # int inl
            ) == 1 or self._handle_openssl_failure()

        else:
            # reference: https://datatracker.ietf.org/doc/html/rfc9001#section-5.4.3

            # AES-based header protected simply samples 16 bytes as input for AES-ECB
            self._binding.lib.EVP_CipherUpdate(
                self._ctx,  # EVP_CIPHER_CTX *ctx
                self._mask,  # unsigned char *out
                self._dummy_outlen,  # int *outl
                self._buffer + sample_offset,  # const unsigned char *in
                SAMPLE_LENGTH,  # int inl
            ) == 1 or self._handle_openssl_failure()

    def _mask_header(self) -> None:
        # use one byte to mask 4 bits for long headers, and 5 bits for short ones
        if self._buffer[0] & 0x80:
            self._buffer[0] ^= self._mask[0] & 0x0F
        else:
            self._buffer[0] ^= self._mask[0] & 0x1F

    def _mask_packet_number(self, pn_offset: int, pn_length: int) -> int:
        # use the remaining (c.f. _mask_header) bytes to mask the packet number field
        # and calculate the truncated packet number
        pn_truncated = 0
        for i in range(pn_length):
            value = self._buffer[pn_offset + i] ^ self._mask[1 + i]
            self._buffer[pn_offset + i] = value
            pn_truncated = value | (pn_truncated << 8)
        return pn_truncated

    def apply(self, plain_header: bytes, protected_payload: bytes) -> bytes:
        # Reference: https://datatracker.ietf.org/doc/html/rfc9001#section-5.4.1
        buffer_len = len(plain_header) + len(protected_payload)
        if buffer_len > PACKET_LENGTH_MAX:
            raise CryptoError("Invalid payload length")

        # read the Packet Number Length from the header
        pn_length = (plain_header[0] & 0x03) + 1

        # the Packet Number is the last field of the header, calculate it's offset
        pn_offset = len(plain_header) - pn_length

        # copy header and payload into the buffer
        self._binding.ffi.memmove(self._buffer, plain_header, len(plain_header))
        self._binding.ffi.memmove(
            self._buffer + len(plain_header), protected_payload, len(protected_payload)
        )

        # build the mask and use it
        self._update_mask(pn_offset, buffer_len)
        self._mask_header()
        self._mask_packet_number(pn_offset, pn_length)

        return self._buffer_view[:buffer_len]

    def remove(self, packet: bytes, encrypted_offset: int) -> Tuple[bytes, int]:
        # Reference: https://datatracker.ietf.org/doc/html/rfc9001#section-5.4.1
        if len(packet) > PACKET_LENGTH_MAX:
            raise CryptoError("Invalid payload length")

        # copy the packet into the buffer
        self._binding.ffi.memmove(self._buffer, packet, len(packet))

        # build the mask and use it to unmask the header first
        self._update_mask(encrypted_offset, len(packet))
        self._mask_header()

        # get the packet number length and unmask it as well
        pn_length = (self._buffer[0] & 0x03) + 1
        pn_truncated = self._mask_packet_number(encrypted_offset, pn_length)

        # return the header and the truncated packet number
        return (
            self._buffer_view[: encrypted_offset + pn_length],
            pn_truncated,
        )



import sys
import os
import platform
from wheel.bdist_wheel import bdist_wheel

import setuptools

include_dirs = [
    os.path.join("vendor", "ls-qpack"),
    os.path.join("vendor", "ls-qpack", "deps", "xxhash"),
]

if sys.platform == "win32":
    extra_compile_args = []
    libraries = [
        "libcrypto",
        "advapi32",
        "crypt32",
        "gdi32",
        "user32",
        "ws2_32",
    ]
    include_dirs.append(
        os.path.join("vendor", "ls-qpack", "wincompat"),
    )
else:
    extra_compile_args = ["-std=c99"]
    libraries = ["crypto"]


class bdist_wheel_abi3(bdist_wheel):
    def get_tag(self):
        python, abi, plat = super().get_tag()

        if python.startswith("cp"):
            return "cp37", "abi3", plat

        return python, abi, plat


if platform.python_implementation() == "CPython":
    extra_kwarg = {
        "py_limited_api": True,
        "define_macros": [("Py_LIMITED_API", "0x03070000")]
    }
else:
    extra_kwarg = dict()


setuptools.setup(
    ext_modules=[
        setuptools.Extension(
            "qh3._buffer",
            extra_compile_args=extra_compile_args,
            sources=["src/qh3/_buffer.c"],
            **extra_kwarg
        ),
        setuptools.Extension(
            "qh3._vendor.pylsqpack._binding",
            extra_compile_args=extra_compile_args,
            include_dirs=include_dirs,
            sources=[
                "src/qh3/_vendor/pylsqpack/binding.c",
                "vendor/ls-qpack/lsqpack.c",
                "vendor/ls-qpack/deps/xxhash/xxhash.c",
            ],
            **extra_kwarg
        ),
    ],
    cmdclass={"bdist_wheel": bdist_wheel_abi3},
)
